"DSA" <- function(Xlearn, Ylearn, binind, maxsize, maxorderint, maxsumofpow, nfolds=5, IDlearn=NULL,
                  forcedterms=NULL, CENSOR=0, weights=NULL, vsplits=NULL, userseed=NULL)
  {
    time0 <- as.POSIXlt(Sys.time())
    
    # set the model fitting debug - if you want to change it comment out this line.
    set.DSA.fit.debug(10)

    checkuserinput(Xlearn,Ylearn,IDlearn,binind,nfolds,userseed,maxsize,maxorderint,
                   maxsumofpow,forcedterms,CENSOR,weights,vsplits)    

    Xlearn <- t(Xlearn)
    if(is.null(IDlearn))IDlearn <- 1:dim(Ylearn)[[1]]
    if(is.null(vsplits))
      {
        #COMPUTE VSPLIT
        if(!is.null(userseed))set.seed(userseed)
        uniqueIDlearn <- t(as.matrix(unique(IDlearn)))
        nIDlearn <- dim(uniqueIDlearn)[[2]]
        #each unique ID is assigned to a split:
        splitvector <- c(rep(1:nfolds,each=floor(nIDlearn/nfolds)),rep(nfolds,nIDlearn-floor(nIDlearn/nfolds)*nfolds)) 
        splitvector <- sample(splitvector)
        #each observation is assigned to the split number to which its
        #corresponding ID was assigned:
        splitvectorlearn <- apply(t(IDlearn),2,function(x,y,z)
                                  {
                                    return(y[z==x])
                                  },y=splitvector,z=uniqueIDlearn)
        vsplits <- apply(t(splitvectorlearn),2,function(x,y)
                         {
                           splitassignment <- rep(0,y)
                           splitassignment[x] <- 1
                           return(splitassignment)
                         },y=nfolds) 
      }
    else userseed <- NULL
    nvarX <- dim(Xlearn)[[1]]
    nlearn <- dim(Xlearn)[[2]]
    maxnval <- max(apply(vsplits,1,sum))
    maxntrain <- nlearn - min(apply(vsplits,1,sum))
    minntrain <- nlearn - max(apply(vsplits,1,sum))

    #create forcedterms if necessary
    if(!is.null(forcedterms))
      {
        Fterms <- labels(terms(forcedterms))
        nFterms <- length(Fterms)
        forcedterms <- matrix(0,nrow=nFterms,ncol=dim(Xlearn)[[1]])
        for(i in 1:nFterms)
          {
            ftermi <- unlist(strsplit(Fterms[i],split="*",fixed=T))
            ftermi <- unlist(apply(cbind(ftermi),1,function(x)return(strsplit(x,split="I(",fixed=T))))
            ftermi <- unlist(apply(cbind(ftermi),1,function(x)return(strsplit(x,split=")",fixed=T))))
            ftermi <- unlist(apply(cbind(ftermi),1,function(x)return(strsplit(x,split=" ",fixed=T))))
            ftermi <- ftermi[ftermi!=""]
            nvari <- length(ftermi)
            powi <- rep(1,nvari)
            for(j in 1:nvari)
              {
                fvari <- unlist(strsplit(ftermi[j],split="^",fixed=T))
                if(length(fvari)==2)
                  {
                    powi[j] <- as.numeric(fvari[2])
                    ftermi[j] <- fvari[1]
                  }
                if(all(ftermi[j]!=dimnames(Xlearn)[[1]]))stop("\nWARNING: Incorrect formula for forcedterms.\n")
                forcedterms[i,ftermi[j]==dimnames(Xlearn)[[1]]] <- powi[j]
              }
          }
      }
    checkuserinput2(maxsize,minntrain)
    
    if(!is.null(forcedterms))nforced <- dim(forcedterms)[1]
    else nforced <- 0
    CVrisks <- matrix(NA,nrow=maxorderint,ncol=maxsize+1)
    PARA <- rep(0,maxsize+1)
    modelselected <- matrix(NA,nrow=maxsize,ncol=nvarX)

    if(CENSOR==0) {
      WTtrainvallearn <- matrix(1, ncol = nlearn, nrow = (nfolds+1))
    }
    else {
      WTtrainvallearn <- weights
    }
    
    gCONST <- c(maxsize,maxorderint,maxsumofpow,nfolds,nlearn,nvarX,CENSOR,NA,NA,binind)

    storage.mode(gCONST) <- "integer"
    storage.mode(Xlearn) <- "double"        
    storage.mode(Ylearn) <- "double"
    storage.mode(CVrisks) <- "double"
    storage.mode(PARA) <- "double"
    storage.mode(vsplits) <- "integer"
    storage.mode(WTtrainvallearn) <- "double"    
    storage.mode(nforced) <- "integer"
    storage.mode(forcedterms) <- "integer"
    storage.mode(maxnval) <- "integer"    
    storage.mode(maxntrain) <- "integer"    
    storage.mode(modelselected) <- "integer"

    if (ncol(Ylearn) == 2) {
      ntrials <- Ylearn[,1] + Ylearn[,2]
      
      # now Ylearn becomes the success ratio.
      Ylearn <- ifelse(ntrials == 0, 0, Ylearn[,1]/ntrials)
      
      # and the weights become the number of trials.
      WTtrainvallearn <- t(apply(WTtrainvallearn, 1, function(row) {
        row * ntrials
      }))
    }
    
    cat("\nEntering the C subroutine.")
    res <- .C("DSA_PACK_Rentry",
              gCONST,
              Xlearn,
              Ylearn,
              CVrisks,
              PARA,
              vsplits,
              WTtrainvallearn,
              nforced,
              forcedterms,
              maxntrain,
              maxnval,
              modelselected,
              NAOK = TRUE
              )
    CVrisks <- res[[4]]
    PARA <- res[[5]]
    gCONST <- res[[1]]
    modelselected <- res[[12]]
    if(binind==0)familyused <- "gaussian"
    else if(binind==1)familyused <- "binomial"
    else familyused <- "unknown"
    modelf <- getmodelf(modelselected,dimnames(Xlearn)[[1]],dimnames(Ylearn)[[2]])
    if(CENSOR==1)weightsused <- weights
    else weightsused <- NULL
    
    final.time <- as.POSIXlt(Sys.time())
    total.time <- difftime(final.time,time0)    
    cat("\nThis call to the DSA routine took ",total.time," ",attributes(total.time)[[2]],".\n",sep="")
    output <- list("average.CVrisks"=CVrisks,"min.risk"=c(gCONST[9],gCONST[8]+1),"model.selected"=modelf,"family"=familyused,"coefficients"=PARA[1:(gCONST[8]+1)],"seed"=userseed,"vsplits"=vsplits,weights=weightsused,"computing.time"=total.time)

    return(output)
  }

"checkuserinput" <- function(Xlearn,Ylearn,IDlearn,binind,nfolds,userseed,maxsize,maxorderint,
                             maxsumofpow,forcedterms,CENSOR,weights,vsplits)
{
  if(!is.matrix(Xlearn))stop("\nWARNING: Xlearn must be a matrix of real numbers.\n")
  if(!is.numeric(Xlearn))stop("\nWARNING: Xlearn must be a matrix of real numbers.\n")
  if(is.null(dimnames(Xlearn)[[2]]))stop("\nWARNING: Xlearn should have a column name.\n")
  if(any(dimnames(Xlearn)[[2]]==""))stop("\nWARNING: Xlearn should have column names.\n")
  if(!all(unique(dimnames(Xlearn)[[2]])==dimnames(Xlearn)[[2]]))
    stop("\nWARNING: the column names of Xlearn should all be different.\n")  

  if (binind) {
    msg.Ylearn <- paste("\nWARNING: Ylearn must be either a one-column matrix of real",
                        "numbers or a two column matrix where the second column is the number of failures.\n")
    if(!is.matrix(Ylearn)) stop(msg.Ylearn)
    if(dim(Ylearn)[[2]] > 2) stop(msg.Ylearn)
    if(!is.numeric(Ylearn)) stop(msg.Ylearn)
  
    if (is.null(colnames(Ylearn)[1]) || colnames(Ylearn)[1] == "")
      stop("\nWARNING: Ylearn should have column name for at least the first column.\n")
  }
  else {
    if(!is.matrix(Ylearn))stop("\nWARNING: Ylearn must be a one-column matrix of real numbers.\n")
    if(dim(Ylearn)[[2]]!=1)stop("\nWARNING: Ylearn must be a one-column matrix of real numbers.\n")
    if(!is.numeric(Ylearn))stop("\nWARNING: Ylearn must be a one-column matrix of real numbers.\n")
    if(is.null(dimnames(Ylearn)[[2]]))stop("\nWARNING: Ylearn should have a column name.\n")
    if(dimnames(Ylearn)[[2]]=="")
      stop("\nWARNING: Ylearn should have a column name.\n")
  }
    
  if(!is.null(IDlearn))if(!is.vector(IDlearn))stop("\nWARNING: IDlearn must be a vector.\n")
  
  if(!is.vector(binind))stop("\nWARNING: binind must be 0 or 1.\n")
  if(length(binind)!=1)stop("\nWARNING: binind must be 0 or 1.\n")
  if( (binind!=0) & (binind!=1) )stop("\nWARNING: binind must be 0 or 1.\n")
  
  if(!is.vector(nfolds))stop("\nWARNING: nfolds must be an integer strictly larger than 1.\n")
  if(length(nfolds)!=1)stop("\nWARNING: nfolds must be an integer strictly larger than 1.\n")
  if(!is.numeric(nfolds))stop("\nWARNING: nfolds must be an integer strictly larger than 1.\n")
  if(nfolds<=1)stop("\nWARNING: nfolds must be an integer strictly larger than 1.\n")  
  if(floor(nfolds)!=nfolds)stop("\nWARNING: nfolds must be an integer strictly larger than 1.\n")

  if( (is.null(vsplits)) & (!is.null(userseed)) )
    {
      if(!is.vector(userseed))stop("\nWARNING: userseed must be an scalar.\n")
      if(length(userseed)!=1)stop("\nWARNING: userseed must be an scalar.\n")
      if(!is.numeric(userseed))stop("\nWARNING: userseed must be an scalar.\n")
    }
  
  if(!is.vector(maxsize))stop("\nWARNING: maxsize must be an integer strictly larger than 0.\n")
  if(length(maxsize)!=1)stop("\nWARNING: maxsize must be an integer strictly larger than 0.\n")
  if(!is.numeric(maxsize))stop("\nWARNING: maxsize must be an integer strictly larger than 0.\n")
  if(maxsize<=0)stop("\nWARNING: maxsize must be an integer strictly larger than 0.\n")
  if(floor(maxsize)!=maxsize)stop("\nWARNING: maxsize must be an integer strictly larger than 0.\n")
  
  if(!is.vector(maxorderint))stop("\nWARNING: maxorderint must be an integer strictly larger than 0.\n")
  if(length(maxorderint)!=1)stop("\nWARNING: maxorderint must be an integer strictly larger than 0.\n")
  if(!is.numeric(maxorderint))stop("\nWARNING: maxorderint must be an integer strictly larger than 0.\n")
  if(maxorderint<=0)stop("\nWARNING: maxorderint must be an integer strictly larger than 0.\n")
  if(floor(maxorderint)!=maxorderint)stop("\nWARNING: maxorderint must be an integer strictly larger than 0.\n")

  if(!is.vector(maxsumofpow))stop("\nWARNING: maxsumofpow must be an integer strictly larger than 0.\n")
  if(length(maxsumofpow)!=1)stop("\nWARNING: maxsumofpow must be an integer strictly larger than 0.\n")
  if(!is.numeric(maxsumofpow))stop("\nWARNING: maxsumofpow must be an integer strictly larger than 0.\n")
  if(maxsumofpow<=0)stop("\nWARNING: maxsumofpow must be an integer strictly larger than 0.\n")
  if(floor(maxsumofpow)!=maxsumofpow)stop("\nWARNING: maxsumofpow must be an integer strictly larger than 0.\n")

  if(!is.null(forcedterms))
    {
      if(as.formula(forcedterms)!=forcedterms)stop("\nWARNING: forcedterms must be a formula.\n")
    }
  
  if(!is.vector(CENSOR))stop("\nWARNING: CENSOR must be 0 or 1.\n")
  if(length(CENSOR)!=1)stop("\nWARNING: CENSOR must be 0 or 1.\n")
  if( (CENSOR!=0) & (CENSOR!=1) )stop("\nWARNING: CENSOR must be 0 or 1.\n")

  if(CENSOR==1)
    {
      if(!is.matrix(weights))stop("\nWARNING: weights must be a matrix of real numbers.\n")
      if(!is.numeric(weights))stop("\nWARNING: weights must be a matrix of real numbers.\n")
    }
  if(!is.null(vsplits))
    {
      if(!is.matrix(vsplits))stop("\nWARNING: vsplits must be a matrix of 0s or 1s.\n")
      if(!is.numeric(vsplits))stop("\nWARNING: vsplits must be a matrix of 0s or 1s.\n")
      if(!all(!is.na(vsplits)))stop("\nWARNING: vsplits must be a matrix of 0s or 1s.\n")
      if(!all(vsplits==1 || vsplits==0))stop("\nWARNING: vsplits must be a matrix of 0s or 1s.\n")
    }

  #CONSISTENCY BETWEEN THE ARGUMENTS

  if(dim(Xlearn)[[1]]!=dim(Ylearn)[[1]])stop("\nWARNING: the number of observations in Xlearn and Ylearn must match.\n")

  if(!all((dimnames(Xlearn)[[2]]!=dimnames(Ylearn)[[2]][1])))
    stop("\nWARNING: the outcome variable should have a name different from all the names of the covariates.\n")

  if (binind == 1 & ncol(Ylearn) == 1) {
    if(!all(is.na(Ylearn[,1]) | Ylearn[,1] == 0 | Ylearn[,1] == 1)) {
      stop("\nWARNING: Ylearn should only contain 0s, 1s or NAs.\m")
    }
  }
  
  if(!is.null(IDlearn)) {
    if (length(IDlearn) != nrow(Ylearn))
      stop("\nWARNING: the length of IDlearn must match the number of observations.\n")
    
    if(length(unique(IDlearn))/nfolds<=1)
      stop(paste("\nWARNING: The number of independent experimental units in the data is",
                 "too low for the requested number of data splits.\n"))
  }
    
  if(CENSOR==1)
    {
      if(dim(weights)[[2]]!=dim(Ylearn)[[1]])
        stop("\nWARNING: the number of columns in weights must match the number of observations.\n")
      if(dim(weights)[[1]]!=(nfolds+1))
        stop("\nWARNING: the number of rows in weights must match the number of data splits.\n")
    }

  if(!is.null(forcedterms))
    {
      if(length(labels(terms(forcedterms)))>maxsize)
        stop("\nWARNING: the number of terms forced in the model should be lower or equal to maxsize.\n")
      if(length(labels(terms(forcedterms)))==maxsize)
        if( (maxorderint!=1) || (maxorderint!=maxsumofpow))
          stop("\nWARNING: If the number of terms forced in the model is equal to maxsize then it is required that maxorderint=maxsumofpow=1.\n")
      if(as.character(forcedterms[[2]])!= colnames(Ylearn)[1])stop("\nWARNING: the outcome name in Ylearn must correspond to the outcome name in the forcedterms.\n")
    }
  if(!is.null(vsplits))
    {
      if(dim(vsplits)[[1]]!=nfolds)stop("\nWARNING: the number of rows of vsplits does not correspond with the value for nfolds.\n")
      if(dim(vsplits)[[2]]!=dim(Ylearn)[[1]])stop("\nWARNING: the number of columns of vsplits does not correspond to the number of observations.\n")
      if(!all(apply(vsplits,1,sum)>0 & apply(vsplits,1,sum)<dim(Ylearn)[[1]]))stop("\nWARNING: the number of observations in each validation sample should be strictly larger than 0 and strictly lower than the total number of observations.")
    }
  if(maxorderint>maxsumofpow)stop("\nWARNING: maxorderint must be lower or equal to maxsumofpow.\n")
}


"checkuserinput2" <- function(maxsize,minntrain)
  {
    if(maxsize+1>minntrain)stop("\nWARNING: The maximum size for the models to be considered is too high for the number of data splits requested and the number of observations/independent units in the data")
  }


"getmodelf" <- function(modelselected,namesxmat,namesymat)
{
  res <- apply(modelselected,1,function(x,y)
               {
                 vars <- y[as.logical(x)]
                 powers <- x[x!=0]
                 termcomponents <- apply(cbind(vars,powers),1,function(x)
                                         {
                                           return(paste(x,collapse="^"))
                                         })
                 term <- ""
                 if(length(termcomponents)>0)term <-
                   paste(" I(",paste(termcomponents,collapse="*"),")",sep="")
                 return(term)
               },y=namesxmat)
  res <- paste(namesymat," ~",paste(res[res!=""],collapse=" +"),sep="")
  return(as.formula(res))
}

"set.DSA.message.level" <- function(newlevel)
  {
    storage.mode(newlevel) <- "integer"
    res <- .C("DSA_PACK_setMessagelevel",newlevel)
    cat("\nThe warning level for the DSA package is set to ",newlevel,".\n")    
  }
