/*****************************************************************************
- runs on learning or training set 
- calls the deletion, substitution, and addition routines
- fits weighted regressions

***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <time.h>

/**
 #include <nag.h>
 #include <nagg02.h>
**/

#include <R.h>
#include <R_ext/Utils.h>
#include "dsa.h"

void DSA_PACK_startAlgo(double *glmWT,double *PARA,double *SE,double *COV,
			double *RES,int *gCOUNT,int *gCONST,double *bestarss,
			double *Ydata,double *Xdata,int *sx,int **bestmodels,int *currentmodel,
			int *tempmodel,int *nextmodel,int *newterm,int *ithterm,double *tempterm,
			double *tempXdata,double *tempYdata,double *WTdata,int trainFlag,
			double *currentarss,int *nforced, int *forcedterms,
			double* new_coefficients,
			double* WX, double* Z, double* qraux, double* effects, double* work, int* pivot)
{
  int IP,msize,maxsize,nvarX,samplesize,nval,nexttoadd,i,j,bestsize,nglmcall,ii,CENSOR;
  int NnonNA,deletion,subsmade,extend,FLAG,binind;
  double oldarss,newarss,temparss,keepBeta;

  /** JHB
      
   Nag_IncludeMean MEAN=Nag_MeanZero;
   Nag_Link LINK=Nag_Logistic;
   Boolean SVD; 
   NagError fail; 

  **/
  
  /* INITIALIZATION */
  IP = 0;
  msize = gCOUNT[0];
  nglmcall = gCOUNT[2];
  nval = gCOUNT[5];

  maxsize = gCONST[0];
  bestsize = gCONST[7];
  nvarX = gCONST[5];
  CENSOR = gCONST[6];
  binind= gCONST[9];

  if(trainFlag == 1)
    samplesize = gCOUNT[4];
  else 
    samplesize = gCONST[4];

  for(i=0;i<(maxsize+1);i++)bestarss[i]=currentarss[i]=R_PosInf;

  if(*nforced==0)
    {
      /* INTERCEPT ONLY MODEL */
      if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)Rprintf("\nFit intercept model");
      IP = 1;
      msize=0;
      DSA_PACK_initialize_l(tempmodel,maxsize*nvarX);
      oldarss=R_PosInf;
      DSA_PACK_evaluatenewmodel(sx,maxsize,msize,nextmodel,-1,tempmodel,nvarX,samplesize,
				tempXdata,Xdata,tempterm,CENSOR,glmWT,WTdata,IP,tempYdata,Ydata,
				PARA,SE,COV,RES,&nglmcall,&oldarss,currentmodel,1,binind,
				new_coefficients, WX, Z, qraux, effects, work, pivot);
      bestarss[IP-1] = oldarss;
      currentarss[IP-1] = oldarss;
    }
  else
    {
      /* FORCED TERMS */
      if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)Rprintf("\nFit forced terms only");
      IP = *nforced+1;
      msize=*nforced;
      DSA_PACK_initialize_l(tempmodel,maxsize*nvarX);
      for(ii=0;ii<*nforced;ii++)for(j=0;j<nvarX;j++)*(tempmodel+ii+j*maxsize)=*(forcedterms+ii+j**nforced);
      oldarss=R_PosInf;
      DSA_PACK_evaluatenewmodel(sx,maxsize,msize,nextmodel,-1,tempmodel,nvarX,samplesize,tempXdata,
				Xdata,tempterm,CENSOR,glmWT,WTdata,IP,tempYdata,Ydata,PARA,SE,COV,
				RES,&nglmcall,&oldarss,currentmodel,1,binind,
				new_coefficients, WX, Z, qraux, effects, work, pivot);
      bestarss[IP-1] = oldarss;
      currentarss[IP-1] = oldarss;
      for(ii=0;ii<IP-1;ii++)
	{
	  bestarss[ii]=R_PosInf;
	  currentarss[ii]=R_PosInf;
	}
    }

  if(*nforced==0)
    {
      /* SELECT BEST ONE-TERM  MODEL */
      if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)Rprintf("\nFit best one-term model");
      newarss=R_PosInf;
      nexttoadd=-1;
      keepBeta=R_PosInf;
      IP=2;
      msize=1;
      for(j=0;j<nvarX;j++)
	{
	  DSA_PACK_initialize_l(tempmodel,maxsize*nvarX);
	  *(tempmodel+j*maxsize)=1;
	  temparss=R_PosInf;
	  DSA_PACK_evaluatenewmodel(sx,maxsize,msize,nextmodel,-1,tempmodel,nvarX,samplesize,
				    tempXdata,Xdata,tempterm,CENSOR,glmWT,WTdata,IP,tempYdata,Ydata,
				    PARA,SE,COV,RES,&nglmcall,&temparss,currentmodel,1,binind,
				    new_coefficients, WX, Z, qraux, effects, work, pivot);
	  if((newarss-temparss)>epsiloncompare)
	    {
	      newarss=temparss;
	      nexttoadd=j;
	      keepBeta = PARA[1];
	    }
	}
      oldarss=newarss;
    }

  /* START DSA - INITIALIZE BESTARSS, BESTINDEXSET, AND CURRINDEXSET */
  if(*nforced==0)
    {
      bestarss[IP-1] = oldarss;
      currentarss[IP-1] = oldarss;
      DSA_PACK_initialize_l(currentmodel,maxsize*nvarX);
      currentmodel[nexttoadd*maxsize+0] = 1;
      DSA_PACK_initialize_l(bestmodels[0],maxsize*nvarX);
      bestmodels[0][nexttoadd*maxsize+0] = 1;
      if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)Rprintf("\nFirst best model: ARSS=%lf",oldarss);
    }
  else
    {
      DSA_PACK_initialize_l(bestmodels[IP-2],maxsize*nvarX);
      for(ii=0;ii<*nforced;ii++)
	for(j=0;j<nvarX;j++)
	  *(bestmodels[IP-2]+ii+j*maxsize)=*(currentmodel+ii+j*maxsize);
      if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)Rprintf("\nFirst model: ARSS=%lf",oldarss);
    }

  if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)Rprintf("\n");
  for(i=0;i<msize;i++)
    {
      if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)Rprintf("%ld \t ",i);
      for(j=0;j<nvarX;j++)
	if(currentmodel[j*maxsize+i] > 0)
	  if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)Rprintf("%ld^%ld ",j,currentmodel[j*maxsize+i]);
      if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)Rprintf("\n");
    }
  if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)Rprintf("\n");

  gCOUNT[0]=msize;
  gCOUNT[2]=nglmcall;
  FLAG = 1;


  if(DSA_PACK_userMlevel>=DSA_PACK_Mlow)Rprintf("\n");
  if(DSA_PACK_userMlevel>=DSA_PACK_Mlow)Rprintf("Start! \n");
  if(DSA_PACK_userMlevel>=DSA_PACK_Mlow)Rprintf("Running algorithm ..... \n");
  if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)Rprintf("\nNumber of terms in model: %li",gCOUNT[0]);
  while(FLAG==1 && gCOUNT[0]<=maxsize && (gCOUNT[0]+1)<samplesize)
    {
      deletion=0;
      subsmade=0;
      extend=1;

      /* D step */  
      if(gCOUNT[0]>max(1,*nforced))
	{
	  if(DSA_PACK_userMlevel>=DSA_PACK_Mhigh)Rprintf("\nDeletion step\n");
	  deletion = DSA_PACK_Dstep(gCOUNT,gCONST,bestarss,bestmodels,tempmodel,currentmodel,
				    nextmodel,newterm,Xdata,tempXdata,Ydata,tempYdata,tempterm,
				    sx,glmWT,PARA,SE,COV,RES,WTdata,trainFlag,currentarss,*nforced,
				    new_coefficients, WX, Z, qraux, effects, work, pivot);

	  if(DSA_PACK_userMlevel>=DSA_PACK_Mhigh)
	    Rprintf("\nNumber of terms in model: %li, deletion=%li",gCOUNT[0],deletion);
	  R_CheckUserInterrupt();
	}
      if(bestarss[gCOUNT[0]] < epsiloncompare)
	{
	  FLAG=0;
	  if(DSA_PACK_userMlevel>=DSA_PACK_Mhigh)
	    Rprintf("\nStop DSA after del. because the last model accepted fits the data very well: please, make sure that the outcome is not in the list of candidate variables.\n");
	}
      /* S step */
      if(deletion==0 && FLAG==1)
	{
	  if(DSA_PACK_userMlevel>=DSA_PACK_Mhigh)Rprintf("\nSubstitution step\n");
	  subsmade = DSA_PACK_Sstep(gCOUNT,gCONST,bestarss,bestmodels,tempmodel,currentmodel,
				    nextmodel,newterm,Xdata,tempXdata,Ydata,tempYdata,tempterm,sx,
				    glmWT,PARA,SE,COV,RES,ithterm,WTdata,trainFlag,currentarss,
				    *nforced, new_coefficients, WX, Z, qraux, effects, work, pivot);
	  if(DSA_PACK_userMlevel>=DSA_PACK_Mhigh)
	    Rprintf("\nNumber of terms in model: %li, subsmade=%li",gCOUNT[0],subsmade);
	  R_CheckUserInterrupt();
	  if(bestarss[gCOUNT[0]] < epsiloncompare)
	    {
	      FLAG=0;
	      if(DSA_PACK_userMlevel>=DSA_PACK_Mhigh)
		Rprintf("\nStop DSA after sub. because the last model accepted fits the data very well: please, make sure that the outcome is not in the list of candidate variables.\n");
	    }
	  /* A step */
	  if(subsmade==0 && FLAG==1)
	    {
	      if(gCOUNT[0]<maxsize)
		{
		  if(DSA_PACK_userMlevel>=DSA_PACK_Mhigh)Rprintf("\nAddition step\n");
		  extend = DSA_PACK_Astep(gCOUNT,gCONST,bestarss,bestmodels,tempmodel,currentmodel,
					  nextmodel,newterm,Xdata,tempXdata,Ydata,tempYdata,tempterm,sx,
					  glmWT,PARA,SE,COV,RES,ithterm,WTdata,trainFlag,
					  currentarss, new_coefficients, WX, Z, qraux, effects, work, pivot);
		  if(DSA_PACK_userMlevel>=DSA_PACK_Mhigh)
		    Rprintf("\nNumber of terms in model: %li, extend=%li",gCOUNT[0],extend);
		  R_CheckUserInterrupt();
		}
	      else if(gCOUNT[0] == maxsize)
		{
		  extend = 0;
		  if(DSA_PACK_userMlevel>=DSA_PACK_Mhigh)
		    Rprintf("\nNot allowed to add anymnore.");
		}
	      if(extend == 0)
		{
		  FLAG=0;
		  if(DSA_PACK_userMlevel>=DSA_PACK_Mhigh)
		    Rprintf("\nStop DSA after add. because the last model accepted fits the data very well: please, make sure that the outcome is not in the list of candidate variables.\n");
		}
	    }
	}
    }    /* end of while loop */
  if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)
    Rprintf("\nAfter DSA in startalgo: globalcount[1]=msize: %li\n\n",gCOUNT[0]);

  if(trainFlag == 0)
    {
      /* OPTIMAL MODEL IS PRINTED TO FORMATOUT */
      IP = bestsize + 1;
      oldarss=R_PosInf;
      

      DSA_PACK_evaluatenewmodel(sx,maxsize,bestsize,tempmodel,-1,bestmodels[bestsize-1],nvarX,
				samplesize,tempXdata,Xdata,tempterm,CENSOR,glmWT,WTdata,IP,tempYdata,
				Ydata,PARA,SE,COV,RES,&nglmcall,&oldarss,currentmodel,1,binind,
				new_coefficients, WX, Z, qraux, effects, work, pivot);
      
      if(DSA_PACK_userMlevel>=DSA_PACK_Mlow)
	{
	  Rprintf("OPTIMAL MODEL: \n");
	  Rprintf("Final ARSS: %lf\n",oldarss);
	  Rprintf("terms in model w intercept %ld \n",IP);
	  Rprintf("terms in model wo intercept %ld \n",bestsize);
	  Rprintf("Estimates of coefficients: \n");
	  DSA_PACK_printmatrix(PARA,IP,1);
	  Rprintf("\n");
	  Rprintf("bestmodels for size %ld \n",bestsize);
	  DSA_PACK_printmatrixl(bestmodels[bestsize-1],maxsize,nvarX);
	  Rprintf("\n");
	  Rprintf("Terms in Final Model: \n");
	  for(ii=0;ii<bestsize;ii++)
	    {
	      Rprintf("%ld \t ",ii);
	      for(i=0;i<nvarX;i++)
		{
		  j = bestmodels[bestsize-1][i*maxsize+ii];
		  if(j > 0)Rprintf("%ld^%ld ",i,j);
		}
	      Rprintf("\n");
	    }
	  Rprintf("\n");
	}


    } /* end of printing if this is DSA on learning set */ 
 

}
